@@ -2,7 +2,7 @@
 
 #' @export
 
-A.l.maker <- function(y, X, Lambda, family, theta_hat, q_l, tps, basehaz) {
+A.l.maker <- function(y, X, Lambda, family, theta_hat, q_l, tps, basehaz, wli) {
   nl <- nrow(X)
   p <- ncol(X)
   if (dim(Lambda)[1]!=dim(Lambda)[2])
@@ -14,7 +14,7 @@ A.l.maker <- function(y, X, Lambda, family, theta_hat, q_l, tps, basehaz) {
     beta_hat <- theta_hat
     power <- matrix(0, len_thet, len_thet)
     for (j in seq_len(nl)) {
-      power <- power + t(X[j, , drop = FALSE]) %*% X[j, , drop = FALSE] *
+      power <- power + t(X[j, , drop = FALSE]) %*% X[j, , drop = FALSE] * wli[j] *
         exp(as.numeric(X[j, ] %*% beta_hat)) /
         (1 + exp(as.numeric(X[j, ] %*% beta_hat)))^2
     }
@@ -29,9 +29,9 @@ A.l.maker <- function(y, X, Lambda, family, theta_hat, q_l, tps, basehaz) {
     power2 <- c(0)
     power3 <- c(rep(0, p))
     for (j in seq_len(nl)) {
-      power <- power + t(X[j, , drop = FALSE]) %*% X[j, , drop = FALSE]
-      power2 <- power2 + (y[j] - as.numeric(X[j, ] %*% beta_hat))^2
-      power3 <- power3 + (y[j] - as.numeric(X[j, ] %*% beta_hat)) *
+      power <- power + t(X[j, , drop = FALSE]) %*% X[j, , drop = FALSE] * wli[j]
+      power2 <- power2 + (y[j] - as.numeric(X[j, ] %*% beta_hat))^2 * wli[j]
+      power3 <- power3 + (y[j] - as.numeric(X[j, ] %*% beta_hat)) * wli[j] *
         t(X[j, , drop = FALSE])
     }
     A_l <- matrix(0, len_thet, len_thet)
@@ -59,7 +59,7 @@ A.l.maker <- function(y, X, Lambda, family, theta_hat, q_l, tps, basehaz) {
       }
     }
     Gamma <- Lambda[1:p,1:p]
-    Gamma_omeg <- Lambda[-(1:p),-(1:p)]
+    #Gamma_omeg <- Lambda[-(1:p),-(1:p)]
     beta_hat <- theta_hat[1:p]
     len_theta <- length(theta_hat)
     if (basehaz %in% c("exp", "gomp", "weibul"))
@@ -73,11 +73,39 @@ A.l.maker <- function(y, X, Lambda, family, theta_hat, q_l, tps, basehaz) {
       B <- i.basis(tps, time, ibasis = TRUE)
       H_est <- as.numeric(B %*% exp(omega_hat))
       Z_beta_hat = as.numeric(Z %*% beta_hat)
-      M_l[1:p, 1:p] <- (t(Z) %*% diag(exp(Z_beta_hat) * H_est) %*% Z )
-      M_l[1:p, (p + 1):len_theta] <- (t(Z) %*% diag(exp(Z_beta_hat))) %*% (B %*% diag(exp(omega_hat)))
+      M_l[1:p, 1:p] <- (t(Z) %*% diag(wli * exp(Z_beta_hat) * H_est) %*% Z)
+      M_l[1:p, (p + 1):len_theta] <- (t(Z) %*% diag(wli * exp(Z_beta_hat))) %*% (B %*% diag(exp(omega_hat)))
       M_l[(p + 1):len_theta, 1:p] <- t(M_l[1:p, (p + 1):len_theta])
-      M_l[(p + 1):len_theta, (p + 1):len_theta] <- diag(as.numeric(exp(Z_beta_hat) %*% B) * exp(omega_hat))
+      M_l[(p + 1):len_theta, (p + 1):len_theta] <- diag(as.numeric((wli * exp(Z_beta_hat)) %*% B) * exp(omega_hat))
       M_l <- M_l/1 + Lambda
+    } else if (basehaz == "unspecified") {
+      ## This matrix can be obtain from optim(hessian=T) in optim.survival()
+      # Initialize the Hessian matrix
+      hessian <- matrix(0, nrow = p, ncol = p)
+      for (i in which(status == 1)) {  # Loop over events only
+        # Risk set: subjects with time >= time[i]
+        risk_set <- which(time >= time[i])
+        # Check if risk_set is non-empty
+        if (length(risk_set) == 0) next
+        # Exponentiated linear predictors for the risk set
+        exp_eta <- exp(Z[risk_set, ] %*% as.matrix(beta_hat))
+        # Calculate the weighted sum of covariates
+        weighted_covariates <- sweep(Z[risk_set, , drop = FALSE], 1, wli[risk_set] * exp_eta, "*")
+        # Weighted average of Z
+        weighted_cov <- colSums(weighted_covariates) / sum(wli[risk_set] * exp_eta)
+        # Weighted covariance matrix of Z in the risk set
+        weighted_cov_matrix <- t(Z[risk_set, , drop = FALSE]) %*%
+          sweep(Z[risk_set, , drop = FALSE], 1, wli[risk_set] * exp_eta, "*") / sum(wli[risk_set] * exp_eta)
+        # Outer product of weighted covariate averages
+        outer_product <- weighted_cov %o% weighted_cov # == weighted_cov %*% t(weighted_cov)
+        # Update the Hessian matrix
+        hessian <- hessian - (weighted_cov_matrix - outer_product)
+      }
+      if (dim(M_l)[1] == dim(hessian)[1]) {
+        M_l <- - hessian + Gamma # not Lambda!
+      } else {
+        stop("Length of 'theta_hat' should be 'p'.") #!
+      }
     } else {
       for (l in 1:len_theta) {
         for (m in 1:len_theta) {
@@ -85,19 +113,19 @@ A.l.maker <- function(y, X, Lambda, family, theta_hat, q_l, tps, basehaz) {
             power <- 0
             if (l <= p & m <= p) {
               for (j in 1:nl) {
-                power <- power +  time[j] * Z[j,][l] * Z[j,][m] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat))
+                power <- power + wli[j] * time[j] * Z[j,][l] * Z[j,][m] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat))
               }
               M_l[l,m] <- power # + Lambda[l,m]
             }
             if (l <= p & m == p+1) {
               for (j in 1:nl) {
-                power <- power +  time[j] * Z[j,][l] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat))
+                power <- power + wli[j] * time[j] * Z[j,][l] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat))
               }
               M_l[l,m] <- M_l[m,l] <- power # + Lambda[l,m]
             }
             if (l == p+1 & m == p+1) {
               for (j in 1:nl) {
-                power <- power +  time[j] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat))
+                power <- power + wli[j] * time[j] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat))
               }
               M_l[l,m] <- power # + Lambda[l,m]
             }
@@ -107,47 +135,48 @@ A.l.maker <- function(y, X, Lambda, family, theta_hat, q_l, tps, basehaz) {
             power <- 0
             if (l <= p & m <= p) {
               for (j in 1:nl) {
-                power <- power + Z[j,][l] * Z[j,][m] * exp(log_a_hat - log_b_hat + as.numeric(Z[j,] %*% beta_hat)) *
+                power <- power + wli[j] * Z[j,][l] * Z[j,][m] * exp(log_a_hat - log_b_hat +
+                                                                      as.numeric(Z[j,] %*% beta_hat)) *
                   (exp(time[j] * exp(log_b_hat)) - 1)
               }
               M_l[l,m] <- power # + Lambda[l,m]
             }
             if (l <= p & m == p+1) {
               for (j in 1:nl) {
-                power <- power +  Z[j,][l] * exp(log_a_hat + time[j] * exp(log_b_hat) +
+                power <- power +  wli[j] * (Z[j,][l] * exp(log_a_hat + time[j] * exp(log_b_hat) +
                                                    as.numeric(Z[j,] %*% beta_hat) - log_b_hat) -
-                  Z[j,][l] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) - log_b_hat)
+                  Z[j,][l] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) - log_b_hat))
               }
               M_l[l,m] <- M_l[m,l] <- power # + Lambda[l,m]
             }
             if (l <= p & m == p+2) {
               for (j in 1:nl) {
-                power <- power +  Z[j,][l] * (time[j] * exp(log_b_hat) -1) *
+                power <- power +  wli[j] * (Z[j,][l] * (time[j] * exp(log_b_hat) -1) *
                   exp(log_a_hat + time[j] * exp(log_b_hat) + as.numeric(Z[j,] %*% beta_hat) - log_b_hat) +
-                  Z[j,][l] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) - log_b_hat)
+                  Z[j,][l] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) - log_b_hat))
               }
               M_l[l,m] <- M_l[m,l] <- power # + Lambda[l,m]
             }
             if (l == p+1 & m == p+2) {
               for (j in 1:nl) {
-                power <- power +  (time[j] * exp(log_b_hat) -1) *
+                power <- power +  wli[j] * ((time[j] * exp(log_b_hat) -1) *
                   exp(log_a_hat + time[j] * exp(log_b_hat) + as.numeric(Z[j,] %*% beta_hat) - log_b_hat) +
-                  exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) - log_b_hat)
+                  exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) - log_b_hat))
               }
               M_l[l,m] <- M_l[m,l] <- power # + Lambda[l,m]
             }
             if (l == p+1 & m == p+1) {
               for (j in 1:nl) {
-                power <- power + exp(log_a_hat + time[j] * exp(log_b_hat) + as.numeric(Z[j,] %*% beta_hat)
-                                     - log_b_hat) - exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) - log_b_hat)
+                power <- power + wli[j] * (exp(log_a_hat + time[j] * exp(log_b_hat) + as.numeric(Z[j,] %*% beta_hat)
+                                     - log_b_hat) - exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) - log_b_hat))
               }
               M_l[l,m] <- power # + Lambda[l,m]
             }
             if (l == p+2 & m == p+2) {
               for (j in 1:nl) {
-                power <- power +  ((time[j] * exp(log_b_hat) )^2 - time[j] * exp(log_b_hat) + 1) *
+                power <- power + wli[j] * (((time[j] * exp(log_b_hat) )^2 - time[j] * exp(log_b_hat) + 1) *
                   exp(log_a_hat + time[j] * exp(log_b_hat) + as.numeric(Z[j,] %*% beta_hat) - log_b_hat) -
-                  status[j] * time[j] * exp(log_b_hat) - exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) - log_b_hat)
+                  status[j] * time[j] * exp(log_b_hat) - exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) - log_b_hat))
               }
               M_l[l,m] <- power # + Lambda[l,m]
             }
@@ -157,44 +186,44 @@ A.l.maker <- function(y, X, Lambda, family, theta_hat, q_l, tps, basehaz) {
             power <- 0
             if (l <= p & m <= p) {
               for (j in 1:nl) {
-                power <- power + Z[j,][l] * Z[j,][m] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) +
+                power <- power + wli[j] * Z[j,][l] * Z[j,][m] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) +
                                                              exp(log_b_hat) * log(time[j]))
               }
               M_l[l,m] <- power # + Lambda[l,m]
             }
             if (l <= p & m == p+1) {
               for (j in 1:nl) {
-                power <- power + Z[j,][l] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) +
+                power <- power + wli[j] * Z[j,][l] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) +
                                                   exp(log_b_hat) * log(time[j]))
               }
               M_l[l,m] <- M_l[m,l] <- power # + Lambda[l,m]
             }
             if (l <= p & m == p+2) {
               for (j in 1:nl) {
-                power <- power + Z[j,][l] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) + log_b_hat +
+                power <- power + wli[j] * Z[j,][l] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) + log_b_hat +
                                                   exp(log_b_hat) * log(time[j])) * log(time[j])
               }
               M_l[l,m] <- M_l[m,l] <- power # + Lambda[l,m]
             }
             if (l == p+1 & m == p+2) {
               for (j in 1:nl) {
-                power <- power +  exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) + log_b_hat +
+                power <- power + wli[j] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) + log_b_hat +
                                         exp(log_b_hat) * log(time[j])) * log(time[j])
               }
               M_l[l,m] <- M_l[m,l] <- power # + Lambda[l,m]
             }
             if (l == p+1 & m == p+1) {
               for (j in 1:nl) {
-                power <- power + exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) +
+                power <- power + wli[j] * exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) +
                                        exp(log_b_hat) * log(time[j]))
               }
               M_l[l,m] <- power # + Lambda[l,m]
             }
             if (l == p+2 & m == p+2) {
               for (j in 1:nl) {
-                power <- power + exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) + log_b_hat +
+                power <- power + wli[j] * (exp(log_a_hat + as.numeric(Z[j,] %*% beta_hat) + log_b_hat +
                                        exp(log_b_hat) * log(time[j])) * log(time[j]) *
-                  (1 + exp(log_b_hat) * log(time[j])) - exp(log_b_hat) * log(time[j]) * status[j]
+                  (1 + exp(log_b_hat) * log(time[j])) - exp(log_b_hat) * log(time[j]) * status[j])
               }
               M_l[l,m] <- power # + Lambda[l,m]
             }
@@ -219,7 +248,7 @@ A.l.maker <- function(y, X, Lambda, family, theta_hat, q_l, tps, basehaz) {
               Lam_0l_omeg_l <- sapply(time, FUN = function(x) integrate(lambda.poly, lower=0,
                                                                         upper=x, p=p, q_l=q_l,
                                                                         beta_dotomega=theta_hat)$value)
-              power <- as.numeric(crossprod(Z[,l] * Z[,m] , Lam_0l_omeg_l * exp(as.numeric(Z %*% beta_hat))))
+              power <- as.numeric(crossprod(Z[,l] * Z[,m] , wli * Lam_0l_omeg_l * exp(as.numeric(Z %*% beta_hat))))
               M_l[l,m] <- power # + Gamma[l,m]
             }
             if (l <= p & m >= p+1) {
@@ -227,7 +256,7 @@ A.l.maker <- function(y, X, Lambda, family, theta_hat, q_l, tps, basehaz) {
                 time, FUN = function(x) integrate(M_k1m2_s_lambda_Taylor, lower=0,
                                                   upper=x, m=m, p=p, q_l=q_l,
                                                   beta_dotomega=theta_hat)$value)
-              power <- as.numeric(crossprod(Z[,l] * s_k1m2_lam_integ, exp(as.numeric(Z %*% beta_hat))))
+              power <- as.numeric(crossprod(Z[,l] * s_k1m2_lam_integ, wli * exp(as.numeric(Z %*% beta_hat))))
               M_l[l,m] <- M_l[m,l] <- power # + Lambda[l,m]
             }
             if (l >= p+1 & m >= p+1) {
@@ -235,7 +264,7 @@ A.l.maker <- function(y, X, Lambda, family, theta_hat, q_l, tps, basehaz) {
                 time, FUN = function(x) integrate(M_k2m2_s_lambda_Taylor, lower=0,
                                                   upper=x, l=l, m=m, p=p, q_l=q_l,
                                                   beta_dotomega=theta_hat)$value)
-              power <- as.numeric(crossprod(s_k2m2_lam_integ, exp(as.numeric(Z %*% beta_hat))))
+              power <- as.numeric(crossprod(s_k2m2_lam_integ, wli * exp(as.numeric(Z %*% beta_hat))))
               M_l[l,m] <- power # + Lambda[l,m]  # or Gamma_omega[l-p,m-p]
             }
             M_l <- M_l/1 + Lambda
@@ -243,8 +272,10 @@ A.l.maker <- function(y, X, Lambda, family, theta_hat, q_l, tps, basehaz) {
         }
       }
     }
-    if (family == "survival" & basehaz == "poly") {
+    if (basehaz == "poly") {
       colnames(M_l) <- rownames(M_l) <- c(colnames(Z), paste("omega",c(0:(len_theta-p-1)), sep="_"))
+    } else if (basehaz == "unspecified") {
+      colnames(M_l) <- rownames(M_l) <- colnames(Z)
     } else {
       colnames(M_l) <- rownames(M_l) <- c(colnames(Z), paste("omega",c(1:(len_theta-p)), sep="_"))
     }
